#include "predict.h"
#include "ray.h"
#include "formulas.h"
	
/***************************************************************************************
							LOAD DATA STRUCTURES SECTION
***************************************************************************************/

	/*********************************************************************************
		Load_Lookups:
			calls routines from FORMULAS.C to load and initialize the necessary
		information for the formulas needed in this model.
			This then allows all formula computations to have immediate access to the
		necessary information they need.
	***********************************************************************************/
void Load_Lookups()
{
	Load_Wall_Coeff();
	Calc_Refl_Coeff(gsRS.iReflTransForm); 
	if (giCL_InBuilding) 	Calc_Trans_Coeff(gsRS.iReflTransForm); 
}

/***************************************************************************************
	 Initialize_Ray_Data_Structures:
	 	- Zeroes the asQP_Store array
		- calls Assign_QP_to_Triangles so that query points are associated with the
			triangles that contain them.  
		- calls Load_Lookups so that ardio formula specific info is loaded for use
			during path-loss calculations
***************************************************************************************/
void Initialize_Ray_Data_Structures(short aiQP_Hit[])
{  
	long	i;

	for(i=0; i<giNumQP; i++) 	aiQP_Hit[i] = 0;
		/* load necessary coefficients, tables and formulas */
	Load_Lookups();
		/* this defines the ground plane */
	gsGroundPlane.sP.fx = (float) 0; gsGroundPlane.sP.fy = (float) 0; gsGroundPlane.sP.fz = (float) 0;
	gsGroundPlane.sN.i = (float) 0; gsGroundPlane.sN.j = (float) 0; gsGroundPlane.sN.k = (float) 1;			   
		/* this defines the ceiling plane */
	gsCeilingPlane.sP.fx = (float) 0; gsCeilingPlane.sP.fy = (float) 0; gsCeilingPlane.sP.fz = gsRS.fCeiling;
	gsCeilingPlane.sN.i = (float) 0; gsCeilingPlane.sN.j = (float) 0; gsCeilingPlane.sN.k = (float) -1;			   
}	

/***************************************************************************************
						INITIALIZE RAY STACK SECTION
***************************************************************************************/

	/*********************************************************************************
		Init_3D_Ray_Stack:
			This algorithm implicitly creates a sphere of radius fCeiling with center
		located at the transmitter location.  Then rays are 'launched' from the sphere 
		according to fAngle.  Necessary info is in pRS.

		'Launched' rays are initialized and then stored in asRayStack.

		All rays to the specified side of the plane, if the plane is defined, are kept.  
				The rest are discarded. iAbove = 0 or 1 (left or right)
	**********************************************************************************/
int Init_3D_Ray_Stack(RAY_TYPE asRayStack[])
{
	int				iNum, i, iCount=0;
	POINT_TYPE		sTransLoc = (gpsTrans_Data+gsRS.iTransID)->sLoc;
	float			fTransPow = (gpsTrans_Data+gsRS.iTransID)->fMax_Power;
	short			j;
	PARA_LINE_TYPE	sLine;
	long			iTriTransIn = PointLocation(sTransLoc, gpObst_Tri, gaiNum_Obst, -1);
	TRIANGLE_TYPE*	psTri = (TRIANGLE_TYPE*) gpObst_Tri[2];
	VERTEX_TYPE		asTriVect[3];

	if (iTriTransIn == -1) { printf("ERROR: transmitter not inside region.\n"); exit(1); }

	iNum = Initialize_Ray(gsRS.iNumRecurse, &gasTriVect);
		/* translate the triangles to the transmitter's location */

	for(i=0; i<iNum; i++) {
		for(j=0; j<3; j++) {
			asTriVect[j].fx = gasTriVect[i][j].fx + sTransLoc.fx;
			asTriVect[j].fy = gasTriVect[i][j].fy + sTransLoc.fy;
			asTriVect[j].fz = gasTriVect[i][j].fz + sTransLoc.fz;
		}
	  	if (giCL_InBuilding || 
	  		!(asTriVect[0].fz >= gsRS.fCeiling - 5 && asTriVect[1].fz >= gsRS.fCeiling - 5 && asTriVect[2].fz + sTransLoc.fz >= gsRS.fCeiling - 5))
	  	{
			asRayStack[iCount].sLKP = ComputeCentroid3D(asTriVect);
			asRayStack[iCount].sVirtTransLoc = sTransLoc;
			sLine = Line_Parametric_Equation3D(sTransLoc, asRayStack[iCount].sLKP);
			asRayStack[iCount].sRayVect = sLine.sV;
			for(j=0; j<3; j++) {
				sLine = Line_Parametric_Equation3D(sTransLoc, asTriVect[j]);
				asRayStack[iCount].asTriVect[j] = sLine.sV;
				asRayStack[iCount].asTriP[j] = asTriVect[j];
			}
			asRayStack[iCount].fAtten = (float) 1;
			asRayStack[iCount].iTriID = PointLocation(asRayStack[iCount].sLKP, gpObst_Tri, gaiNum_Obst, (psTri+iTriTransIn)->iEdge[0]);
			asRayStack[iCount].iEdgeID = From_Edge(asRayStack[iCount], asRayStack[iCount].sLKP);
			asRayStack[iCount].cNumEvent = '0';
			asRayStack[iCount].cPrevious = '0';
			for(j=0; j<MAX_EVENTS; j++) {
				asRayStack[iCount].aiPath[j] = -1;	
				asRayStack[iCount].acEvent[j] = '0';
			}
			iCount++;
		}		
	}
	if (giCL_RaySave) { giInitRayNum = iCount; }

	printf("   Init Rays: %i	Lower: %i\n", iNum, iCount); 

	return(iCount);
}
/* int Diffraction_Update(POINT_TYPE sDiffLoc, RAY_TYPE* psCurrentRay)
{
}  */
	/*********************************************************************************
		Initialize_Ray_Stack:
			Calls the appropriate routine to setup the ray stack.  The angle determines
		the separation between rays.

	 - Returns the number of rays created and stored in the ray stack.
	 - sets gpsRayStack to the top of the stack
	**********************************************************************************/
int Initialize_Ray_Stack(RAY_TYPE asRayStack[])
{
	int			iNum;

	if (gi3DOn) 
		iNum = Init_3D_Ray_Stack(asRayStack);

	gpsRayStack = &asRayStack[iNum - 1];

	return(iNum);
}

/***************************************************************************************
								RAY WALK SECTION
***************************************************************************************/

	/*********************************************************************************
		Compute_Trap: this function calculates the trapezoid used for point capture
			it computes the 5 face planes and points their normals into the trapezoid
	**********************************************************************************/
TRAP_TYPE Compute_Trap(POINT_TYPE sOrigP[], POINT_TYPE sFinalP[], short iNumEvents)
{
	TRAP_TYPE		sTrap;
	short			i;
	float			t;

		/* compute face normals */
			/* sides */
	sTrap.sP[0] = Plane_Point_Normal_Plus(sOrigP[0], sOrigP[1], sFinalP[0], sFinalP[2]);
	sTrap.sP[1] = Plane_Point_Normal_Plus(sOrigP[1], sOrigP[2], sFinalP[1], sFinalP[0]);
	sTrap.sP[2] = Plane_Point_Normal_Plus(sOrigP[0], sOrigP[2], sFinalP[0], sFinalP[1]);
		/* small end face */
	sTrap.sP[3] = Plane_Point_Normal_Plus(sOrigP[0], sOrigP[1], sOrigP[2], sFinalP[2]);
		/* large end face */
	sTrap.sP[4] = Plane_Point_Normal_Plus(sFinalP[0], sFinalP[1], sFinalP[2], sOrigP[2]);
	
		/* now we need to slightly enlarge the trapezoid to account for round off errors */
	t = (float) -.05 * (float) iNumEvents;
	for(i=0; i<5; i++)
		sTrap.sP[i].sP = ParaLine_Calc_Point(sTrap.sP[i].sN, sTrap.sP[i].sP, t); 

	return(sTrap);
}

PLANE_TYPE	Create_Edge_Plane(long iEdgeID, POINT_TYPE sPdir)
{
	POINT_TYPE		sP[2];
	VERTEX_TYPE*	psVertices = (VERTEX_TYPE*) gpObst_Tri[0];
	EDGE_TYPE*		psEdges = (EDGE_TYPE*) gpObst_Tri[1];
	PLANE_TYPE		sPlane;
	short			i;
	VECTOR_TYPE		sV;
	float			fDiff;

	   	/* assumes that the plane is perpendicular to the xy plane */
	for(i=0; i<2; i++)
		sP[i] = psVertices[(psEdges+iEdgeID)->iPoint[i]];

	sV.k = (float) fabs(0); 
	sV.j = (float) 1;
	fDiff = sP[1].fx - sP[0].fx;
	if (Macro_NotZero(fDiff)) {
		sV.i = (sP[0].fy - sP[1].fy) / fDiff;
		sV = Normalize_Vector(sV);
	}
	else
		{ sV.i = (float) 1; sV.j = (float) 0; }
				/* virtual transmitter location is 'above' the edge-plane */
	if (Point_To_Plane_DistancePM2(sV, sP[0], sPdir) < 0)
			{ sV = Vector_MultBy_Scaler(sV, (float) -1); }

	sPlane.sN = sV;
	sPlane.sP = sP[0];

	return(sPlane);
}
	/*********************************************************************************
		Ray_TriEdge_Intersect:	
			Finds the next edge the ray-cone intersects with and then returns the
		Triangle ID associated with it.  Used to walk the ray-cone through the
		triangulation.
			*piEdge = the edge ID or -2 or -3 if the ray only ground/ceiling intersects
				(if the .i and .j components of the RayVect are very small then only 
															ground or ceiling reflects)
			*piEdge = -1, then ray is leaving the bounding region
			*piEdge = -4 then ray must be aborted since the proper edge for 
					intersection could not be found
	**********************************************************************************/
long Ray_TriEdge_Intersect(RAY_TYPE* pCurrRay, long* piEdge)
{
	long			iTriID, iEdgeID;
	POINT_TYPE		sRayPoint1, sRayPoint2, sEdgePoint1, sEdgePoint2;
	VERTEX_TYPE*	psVertices = (VERTEX_TYPE*) gpObst_Tri[0];
	EDGE_TYPE*		psEdges = (EDGE_TYPE*) gpObst_Tri[1];
	TRIANGLE_TYPE*	psTri = (TRIANGLE_TYPE*) gpObst_Tri[2];
	short			i, Inter, iTries=0;
	VECTOR_TYPE		sV;

	if (giCL_Debug) {
		fprintf(gfp_Debug, "  RAY_TRI_EDGE_INTERSECT:\n");
		fprintf(gfp_Debug, "    iTri %li %li\n    ", pCurrRay->iTriID, pCurrRay->iEdgeID); 
		fPrint_Vector(gfp_Debug, pCurrRay->sRayVect);
		fprintf(gfp_Debug, "    VirtTrans: "); fPrint_Point(gfp_Debug, pCurrRay->sVirtTransLoc); 
		fprintf(gfp_Debug, "    LKP:       "); fPrint_Point(gfp_Debug, pCurrRay->sLKP);
		SaveDF();
	}

	iTriID = pCurrRay->iTriID;
	sRayPoint1 = pCurrRay->sLKP;
	sV = pCurrRay->sRayVect;
Again:
	sRayPoint2 = ParaLine_Calc_Point(sV, sRayPoint1, (float) 1000);

	i = 0; Inter = 0;
	while (i<3)
	{
		iEdgeID = (psTri+iTriID)->iEdge[i];
		if (iEdgeID != pCurrRay->iEdgeID)
		{
			sEdgePoint1 = psVertices[(psEdges+iEdgeID)->iPoint[0]];
			sEdgePoint2 = psVertices[(psEdges+iEdgeID)->iPoint[1]];
			if (Intersect_Segments2D(sRayPoint1, sRayPoint2, sEdgePoint1, sEdgePoint2))
				{ i = 3; Inter = 1; }
		}
		i++;
	}
	if (!Inter)  {			
				/* only ground reflects or ceiling reflects */
		if (pCurrRay->sRayVect.i < .05 && pCurrRay->sRayVect.j < .05) {
			if (pCurrRay->sRayVect.k < 0)
				*piEdge = -2; /* for ground */
			else
				*piEdge = -3; /* for ceiling */
			return(pCurrRay->iTriID);
		}
		else { /* must have hit a vertex or ray-cone believes it is in the wrong triangle */
			if (iTries == 0) {	
				iTriID = PointLocation(pCurrRay->sLKP, gpObst_Tri, gaiNum_Obst, -1);
				giNumInter++;
				if (iTriID == -1)	return(-1);		/* outside of bounding region */
			}
			if (pCurrRay->iTriID != iTriID) {
				if (giCL_Debug)	 fprintf(gfp_Debug, "         Ray_TriEdge: RayTriID != iTriID\n");
		/* printf("RayInter: iTriIDs don't match\n"); */
				pCurrRay->iTriID = iTriID; 
				pCurrRay->iEdgeID = From_Edge(*pCurrRay, pCurrRay->sLKP);
			}
				/* fudge the vector a little to solve the vertex hit problem */
			sV.i = sV.i + (float) .001; sV.j = sV.j + (float) .001;
			iTries++;
			if (iTries < 3 && pCurrRay->iTriID != -1) 	goto Again; 
			else {
				if (giCL_Debug)	 fprintf(gfp_Debug, "        Ray is Aborted at Ray_Tri 1\n");
				return(-4); /* abort ray after 3 tries */
			}
		}
	}
	*piEdge = iEdgeID;
	
	iTriID = OtherTri(iTriID, iEdgeID);
		/* if OtherTri returns -2 it means that iEdgeID does not belong to iTriID */
	if (iTriID == -2)		{ printf("OtherTri == -2: RayEdge\n"); return(-4); }

	return(iTriID); 
}
	/*********************************************************************************
		Compute_Cone_End_Inter:
			computes the intersection of the cone with 
		if (iEdgeID > 0)	EdgePlane
			iEdgeID = -2	Ground
			iEdgeID = -3	Ceiling
	**********************************************************************************/
void Compute_Cone_End_Inter(RAY_TYPE* pCurrRay, long iEdgeID, POINT_TYPE asInter[])
{
	PLANE_TYPE		sPlane;
	PARA_LINE_TYPE	sLine;
	int				i;

	if (iEdgeID == -2)  /* ground plane */
		if (giCL_Terrain)	
			sPlane = *(gpsTerrPlanes + pCurrRay->iTriID);  
		else	
			sPlane = gsGroundPlane;
	else if (iEdgeID == -3)
		sPlane = gsCeilingPlane;
	else
		sPlane = Create_Edge_Plane(iEdgeID, pCurrRay->sVirtTransLoc);
	
	sLine.sP = pCurrRay->sVirtTransLoc; 		
	sLine.sV = pCurrRay->sRayVect;
	asInter[3] = Line_Plane_Intersection(sLine, sPlane);
	for(i=0; i<3; i++) {
		sLine.sV = pCurrRay->asTriVect[i];
		asInter[i] = Line_Plane_Intersection(sLine, sPlane);
	}
}
	/*********************************************************************************
		Event_Check:
			This function checks for events and updates the ray stack accordingly;

		As a ray 'walks' through the triangulation these events are checked for:
		0 - intersection with a wall
		1 - intersection with a vertex or edge != iEdgeID of the triangulation
		2 - intersection with the ground
		3 - intersection with the ceiling

		asInter[]: contains the intersection points of the ray-cone with the edgeplane,
			ceilingplane or groundplane (asInter[3] is the center ray's intersection point)
	**********************************************************************************/
short Event_Check(short aiEvents[], RAY_TYPE* pCurrRay, long iNextTri, long* piEdgeID,
												POINT_TYPE asInter[], long aiStatus[])
{
	short			i, j, iEvent=0, iAllG=0, iAllC=0;
	VERTEX_TYPE*	psVertices = (VERTEX_TYPE*) gpObst_Tri[0];
	EDGE_TYPE*		psEdges = (EDGE_TYPE*) gpObst_Tri[1];
	TRIANGLE_TYPE*	psTri = (TRIANGLE_TYPE*) gpObst_Tri[2];
	PLANE_TYPE		sPlane;
	long			iTriID;
	
			/* sometimes the artificially created boundary is negative but
					it is assumed that the environment is always positive */
	if (asInter[3].fx < 0 || asInter[3].fy < 0)
		{ aiEvents[6]=1; iEvent=1;  return(iEvent); }

		/* ray intersects a wall */
	if (*piEdgeID >= 0 && (psEdges+*piEdgeID)->iWallType > 0)	
		{ aiEvents[0] = 1; iEvent = 1; }

		/* check for ground reflection */
	if (pCurrRay->cPrevious != '3') {	/* this handles floating point error problems */
		if (*piEdgeID == -2) {
			aiEvents[2] = 1; iEvent = 1; 
			for(i=0; i<3; i++)	{ aiStatus[i] = -2; }
		}
		else if (*piEdgeID >= 0) {
			if (giCL_Terrain) { /* check to see if any points lie below the edge */
				sPlane = *(gpsTerrPlanes + pCurrRay->iTriID);
				for(i=0; i<3; i++)
					if (Point_To_Plane_DistancePM(sPlane, asInter[i]) < -0)	
						{ aiStatus[i] = -2; aiEvents[2] = 1; iEvent = 1; }
			}
			else
				for(i=0; i<3; i++)
					if (asInter[i].fz < -0)	
						{ aiStatus[i] = -2; aiEvents[2] = 1; iEvent = 1; }
		}
	}

		/* check for a ceiling reflection */
	if (pCurrRay->cPrevious != '4') {	/* this handles floating point error problems */
		if (*piEdgeID == -3) { 
			aiEvents[3] = 1; iEvent = 1; 
			for(i=0; i<3;i++)	aiStatus[i] = -3;
		}
		else if (*piEdgeID >= 0) 
			for(i=0; i<3; i++)	
				if (asInter[i].fz > gsRS.fCeiling)	
					{ aiStatus[i] = -3; aiEvents[3] = 1; iEvent = 1; }
	}

		/* check to see if the ray-cone needs to be split */
	for(i=0; i<2; i++)
		for(j=i; j<3; j++)
			if (aiStatus[i] != aiStatus[j])
				{ aiEvents[1] = 1; iEvent = 1; i=2; j=3; }

		/* check to see if it only ground or ceiling reflects */
	if (aiEvents[1] == 0 && (aiEvents[2]==1 || aiEvents[3]==1)) {
		iAllG = 1; iAllC = 1;
		for(j=0; j<3; j++) {
			if (aiStatus[j] != -2)	iAllG = 0;
			if (aiStatus[j] != -3)	iAllC = 0;
		}
		if (iAllG) { *piEdgeID = -2; Compute_Cone_End_Inter(pCurrRay, -2, asInter); aiEvents[0]==0; }
		if (iAllC) { *piEdgeID = -3; Compute_Cone_End_Inter(pCurrRay, -3, asInter); aiEvents[0]==0; }
	}
	
	
		/* check if ray-cone end face lies in more than one triangle */
	for(i=0; i<3; i++) {
	  iTriID = PointLocation(asInter[i], gpObst_Tri, gaiNum_Obst, (psTri + pCurrRay->iTriID)->iEdge[0]); 
	
	if ((iTriID != pCurrRay->iTriID && *piEdgeID >= 0 && iTriID != iNextTri) ||
			(iTriID != pCurrRay->iTriID && *piEdgeID < 0)) 
		{
			if (iTriID < 0)	aiStatus[i] = iTriID; else aiStatus[i] = iTriID + 1;
			aiEvents[1] = 1; iEvent = 1; 
		}
	}
	return(iEvent);
}

	/*********************************************************************************
		Reflect: reflects a ray off the plane

		returns the new virtual transmitter location and the reflection vector
	**********************************************************************************/
VECTOR_TYPE	Reflect(PARA_LINE_TYPE sRay, PLANE_TYPE sPlane, POINT_TYPE* pNewVirtLoc,
																	POINT_TYPE	sInter)
{
	PARA_LINE_TYPE	sLine;
	POINT_TYPE		sNewLoc, sPtemp;
	UNIT_VECTOR		sNewVectDir, sV;
	float			fDist, fTemp;

		/* setup the line for computing the next virtual transmitter location */
	sLine.sP = sRay.sP; sLine.sV = sPlane.sN;
		/* compute the distance from the Virtual transmitter to the plane */
		/* -2 because the plane normal points towards the virtual transmitter */
	fDist = -2 * Point_To_Plane_Distance(sPlane, sRay.sP);
		/* for test purposes */
	if (giCL_Debug) {
		sPtemp = Line_Plane_Intersection(sLine, sPlane);
		fTemp = -2 * Macro_SqrtDist(sLine.sP, sPtemp);
		if (fTemp > fDist + .01 || fTemp < fDist - .01) {
			fprintf(gfp_Debug, "   REFLECT: Reflection Distances don't match: %f %f\n", fTemp, fDist); 
			SaveDF(); }
	}
		/* compute the new virtual transmitter location */
	sNewLoc = ParaLine_Calc_Point(sLine.sV, sLine.sP, fDist);
		/* compute the reflection vector */
	sLine = Line_Parametric_Equation3D(sNewLoc, sInter);
	sNewVectDir = sLine.sV;
		/* for test purposes */
	if (giCL_Debug) {
		sV = Reflection_Vector(sRay.sV, sPlane);
		if (!(Macro_FloatEqual(sNewVectDir.i, sV.i) && Macro_FloatEqual(sNewVectDir.j, sV.j) && Macro_FloatEqual(sNewVectDir.k, sV.k))) {
			fprintf(gfp_Debug, " >>REFLECT: vectors do not match\n    NewVirtLoc: "); fPrint_Point(gfp_Debug, sNewLoc); 
			fprintf(gfp_Debug, "   IntersectionPoint: "); fPrint_Point(gfp_Debug, sInter);
			fprintf(gfp_Debug, "   NewVect: "); fPrint_Vector(gfp_Debug, sNewVectDir);
			fprintf(gfp_Debug, "   CheckVect: "); fPrint_Vector(gfp_Debug, sV);
		}
 		if (Macro_Zero(sNewVectDir.k)) fprintf(gfp_Debug, " >>REFLECT: .k == 0\n");
		SaveDF();
	}
	/* */

	*pNewVirtLoc = sNewLoc;
	return(sNewVectDir);
}
	/************************************************************************
		Reflect_Ray: omputes the new virtual transmitter location using Reflect
			then computes the new ray-cone vectors

			asInter[]: the cone's intersection points on the plane sPlane
				asInter[3] is the centroid
	************************************************************************/
void Reflect_Ray(RAY_TYPE* pCurrRay, POINT_TYPE asInter[], PLANE_TYPE sPlane)
{
	PARA_LINE_TYPE	sRay, sLine;
	POINT_TYPE		sNewVert;
	VECTOR_TYPE		sNewVect;
	short			i;

	sRay = Line_Parametric_Equation3D(pCurrRay->sVirtTransLoc, asInter[3]);
	sNewVect = Reflect(sRay, sPlane, &sNewVert, asInter[3]);
		/* update Ray */
	pCurrRay->sRayVect = sNewVect;
	pCurrRay->sVirtTransLoc = sNewVert;
	for(i=0; i<3; i++) { 
		sLine = Line_Parametric_Equation3D(sNewVert, asInter[i]);
		pCurrRay->asTriVect[i] = sLine.sV;
	}
}

	/**********************************************************************************
	**********************************************************************************/
void Break_Up_Ray_Edge(RAY_TYPE* pCurrRay, POINT_TYPE asInter[], PLANE_TYPE	sPlane1,
						PLANE_TYPE sPlane2, POINT_TYPE sP1, POINT_TYPE sP2, short iPlane,
						RAY_TRI_TYPE asNewTri[]) 
{
	POINT_TYPE		sInter[2], sPl[2][3], asRotInter[4], sRotInter[2], sP1R, sP2R, sPtemp;
	short			i=0, iC[2], j, k, iFlag, iRot;
	PARA_LINE_TYPE	sLine, sL1, sL2;
	float			fCosTheta, fSinTheta, t;

	iC[0]=0;
		/* angle sPlane1 makes with the xy plane */
	fCosTheta = Macro_DotProduct(sPlane1.sN, gsGroundPlane.sN); 
/*	if (sP1.fz > 0 || sP2.fz > 0)
	else fCosTheta = 1; */
	fSinTheta = (float) sqrt(1 - pow(fCosTheta, 2));
/* printf(">>>>>> Break_Up_Ray_Edge  %f  %f\n", fCosTheta, fSinTheta); */

			/* set whether to rotate through X or Y */
	if (Macro_FloatEqual(asInter[0].fx, asInter[1].fx) && Macro_FloatEqual(asInter[1].fx, asInter[2].fx)) 
		iRot = 1; else iRot = 0;
	sL1 = Line_Parametric_Equation3D(sP1, sP2); 

	for(j=0; j<4; j++)	asRotInter[j] = Rotate_CosXYZ(asInter[j], fCosTheta, fSinTheta, iRot);
	sP1R = Rotate_CosXYZ(sP1, fCosTheta, fSinTheta, iRot);
	sP2R = Rotate_CosXYZ(sP2, fCosTheta, fSinTheta, iRot);

		/* now we are working on the xy plane */
	if (Intersect_Segments2D(asRotInter[0], asRotInter[1], sP1R, sP2R)) {
		sRotInter[i] = Calc_Intersection_Of_Segments2D(asRotInter[0], asRotInter[1], sP1R, sP2R, &iFlag);
if (iFlag == 0) sL2 = Line_Parametric_Equation3D(asInter[0], asInter[1]);
		if (iFlag == 0) 		i++;
	}
	if (Intersect_Segments2D(asRotInter[0], asRotInter[2], sP1R, sP2R)) {
		sRotInter[i] = Calc_Intersection_Of_Segments2D(asRotInter[0], asRotInter[2], sP1R, sP2R, &iFlag);
if (i==0 && iFlag == 0) sL2 = Line_Parametric_Equation3D(asInter[0], asInter[2]);
		if (iFlag == 0)		i++;
	}
	if (i < 2 && Intersect_Segments2D(asRotInter[1], asRotInter[2], sP1R, sP2R)) {
		sRotInter[i] = Calc_Intersection_Of_Segments2D(asRotInter[1], asRotInter[2], sP1R, sP2R, &iFlag);
		if (iFlag == 0) 	i++;
	}
abort:
	if (i<2)	{ 
	/*	printf("Break_Up: %i Intersection point(s)\n", i); 
		Print_Ray(*pCurrRay); 
		Print_Point(asInter[0]); Print_Point(asInter[1]); Print_Point(asInter[2]); Print_Point(asInter[3]); 
		Print_Point(sP1); Print_Point(sP2);
		Print_Point(asRotInter[0]); Print_Point(asRotInter[1]); Print_Point(asRotInter[2]); Print_Point(asRotInter[3]); 
		Print_Point(sP1R); Print_Point(sP2R); */
		for(i=0; i<4; i++)	asNewTri[0].asP[i] = asInter[i];
		asNewTri[0].iPlane = 0;
		asNewTri[0].iActive = 1;  
		asNewTri[1].iActive = 0;  
		asNewTri[2].iActive = 0;  
		/* printf("  >>>>>>>>>> done %i\n", giNumRayStack); */
		return; 
	}

	sInter[0] = Rotate_CosXYZ(sRotInter[0], (float) -1 * fCosTheta, (float) -1 * fSinTheta, iRot);
	sInter[1] = Rotate_CosXYZ(sRotInter[1], (float) -1 * fCosTheta, (float) -1 * fSinTheta, iRot);

	if (Macro_NotZero(sL1.sV.k)) t = (sInter[0].fz - sL1.sP.fz) / sL1.sV.k;
	else if (Macro_NotZero(sL1.sV.i)) t = (sInter[0].fx - sL1.sP.fx) / sL1.sV.i;
	else t = (sInter[0].fy - sL1.sP.fy) / sL1.sV.j;
	if (iRot == 0) 	sInter[0].fy = sL1.sP.fy + t * sL1.sV.j;
	else 	sInter[0].fx = sL1.sP.fx + t * sL1.sV.i;
	if (Macro_NotZero(sL1.sV.k)) t = (sInter[1].fz - sL1.sP.fz) / sL1.sV.k;
	else if (Macro_NotZero(sL1.sV.i)) t = (sInter[1].fx - sL1.sP.fx) / sL1.sV.i;
	else t = (sInter[1].fy - sL1.sP.fy) / sL1.sV.j;
	if (iRot == 0)	sInter[1].fy = sL1.sP.fy + t * sL1.sV.j;
	else sInter[1].fx = sL1.sP.fx + t * sL1.sV.i;

	sInter[0] = Nearest_Point_On_Plane(sPlane1, sInter[0]);
	sInter[1] = Nearest_Point_On_Plane(sPlane1, sInter[1]);

/********/
		/* calculate new intersection points for tri points that need to be on Plane2 */
	iC[0]=0; iC[1]=0;
	for(i=0; i<3; i++) {
		sLine = Line_Parametric_Equation3D(pCurrRay->sVirtTransLoc, asInter[i]);
		sPtemp = Line_Plane_Intersection(sLine, sPlane2);
		if (Macro_Dist(sPtemp, pCurrRay->sVirtTransLoc) < Macro_Dist(asInter[i], pCurrRay->sVirtTransLoc)) {
			sPl[1][iC[1]] = sPtemp;
			iC[1]++;
		}
		else {
			sPl[0][iC[0]] = asInter[i]; 
			iC[0]++;
		}
	}
if (iC[0] == 3)	{ /* printf("->>>>>>>>>>>>>>>>>>>>>>> iC[0] ==3 ");*/ i=0; goto abort; }
if (iC[1] == 3)	{ /* printf("->>>>>>>>>>>>>>>>>>>>>>> iC[1] ==3 "); */ i=0; goto abort; }
	/*******/
		/* now setup each triangle.  Need to split the trapezoid and shift the inter points */
	j=0;
	for(k=0; k<2; k++) {
		if (iC[k] == 2) {
			asNewTri[j].asP[0] = sPl[k][0]; asNewTri[j].asP[1] = sPl[k][1]; asNewTri[j].asP[2] = Point_Offset(sInter[0], sPl[k][0]);
			asNewTri[j].asP[3] = ComputeCentroid3D(asNewTri[j].asP);
			asNewTri[j].iPlane = k; asNewTri[j].iActive = 1;
			j++;
			if (Intersect_Segments2D(Rotate_CosXYZ(sPl[k][0], fCosTheta, fSinTheta, iRot), sRotInter[1], Rotate_CosXYZ(sPl[k][1], fCosTheta, fSinTheta, iRot), sRotInter[0]))
				{ asNewTri[j].asP[0] = sPl[k][1]; asNewTri[j].asP[1] = Point_Offset(sInter[0], sPl[k][1]); asNewTri[j].asP[2] = Point_Offset(sInter[1], sPl[k][1]); }
			else
				{ asNewTri[j].asP[0] = sPl[k][0]; asNewTri[j].asP[1] = Point_Offset(sInter[0], sPl[k][0]); asNewTri[j].asP[2] = Point_Offset(sInter[1], sPl[k][0]); }
		}
		else {
			asNewTri[j].asP[0] = sPl[k][0]; asNewTri[j].asP[1] = Point_Offset(sInter[0], sPl[k][0]); asNewTri[j].asP[2] = Point_Offset(sInter[1], sPl[k][0]);
		}
		asNewTri[j].asP[3] = ComputeCentroid3D(asNewTri[j].asP);
		asNewTri[j].iPlane = k; asNewTri[j].iActive = 1;
		j++;
	}
}
	/**********************************************************************************
		Break_Up_Ray: this routine splits the current ray-cone into as many
			new ray-cones as necessary so that all new ray-cones will lie within
			a single triangle
		
		  This requires intersecting the ray-cone with the prism defined by the 
		current triangle.  For speed, rather than use a brute force intersection
		computation the method applied here is to split the ray-cone across edges
		until no more edges exist that need to be split across.

		  It returns a list of the triangle points incident on th edge, ground or
		ceiling planes as required.  
	**********************************************************************************/
void Break_Up_Ray(RAY_TYPE* pCurrRay, long iEdgeID, PLANE_TYPE sInterPlane, 
							POINT_TYPE asInter[], long aiStatus[], RAY_TRI_TYPE asNewTri[], int iLevel)
{
	short			i, j, iYes=0, iYes2=0, k, iCount, iNewCount, iAll;
	long			iEdgeID2, aiStat[4];
	POINT_TYPE		sP1, sP2, asP[4];
	PLANE_TYPE		sPlane, sPlane2;

	gpsVert = (VERTEX_TYPE*) gpObst_Tri[0]; gpsEdge = (EDGE_TYPE*) gpObst_Tri[1]; 
/* printf("BREAK_UP_RAY: %li %li %li %li\n", iEdgeID, aiStatus[0], aiStatus[1], aiStatus[2]); */
	if (iEdgeID > 0) {	/* normal case where sRayVect has intersected an edge */
		for(i=0; i<3; i++) {
				/* > 0 means it has the Triangle ID stored in it */
			if (aiStatus[i] > 0 || aiStatus[i] == -1) {
/* printf("Break1: normal \n"); */
				iEdgeID2 = Other_Edge(pCurrRay->iTriID, iEdgeID, pCurrRay->iEdgeID);
				sP1 = Common_Edge_Vertex(iEdgeID, iEdgeID2); sP1.fz = (float) 0;
				sP2 = sP1;  sP2.fz = (float) 1000; sP1.fz = (float) -1000;
				sPlane2 = Create_Edge_Plane(iEdgeID2, pCurrRay->sVirtTransLoc);
					/********/
				Break_Up_Ray_Edge(pCurrRay, asInter, sInterPlane, sPlane2, sP1, sP2, 1, asNewTri);
					/********/
				i=3; iYes=1;
			}
		}
		if (iYes==0) {	/* didn't need to split across edges, test floor and ceiling */
			for(i=0; i<3; i++) {
					/* < 0 means point is below (above) the ground (ceiling) (-2 and -3) */
				if (aiStatus[i] < 0) {
/* printf("Break2: normal G or C\n"); */
					sP1 = *(gpsVert + (gpsEdge+iEdgeID)->iPoint[0]);
					sP2 = *(gpsVert + (gpsEdge+iEdgeID)->iPoint[1]);
					if (aiStatus[i] == -2) /* ground */
						if (giCL_Terrain) 	sPlane2 = *(gpsTerrPlanes+pCurrRay->iTriID);
						else	sPlane2 = gsGroundPlane;
					else {
						sPlane2 = gsCeilingPlane;
						sP1.fz = gsRS.fCeiling;  sP2.fz = gsRS.fCeiling;
					}
					Extend_Points(&sP1, &sP2);
						/********/
					Break_Up_Ray_Edge(pCurrRay, asInter, sInterPlane, sPlane2, sP1, sP2, -1, asNewTri);
						/********/
					iEdgeID2 = aiStatus[i];
					i=3;
				}
			}
		}
	}
	else {	/* we are working with the ground or ceiling plane */
/* printf("Break3: G or C\n"); */
		for(i=0; i<3; i++) {
			if (aiStatus[i] > 0 || aiStatus[i] == -1) {
				iEdgeID2 = Tri_Edge_Segment_Inter(pCurrRay->iTriID, asInter[3], asInter[i]);
				sP1 = *(gpsVert + (gpsEdge+iEdgeID2)->iPoint[0]);
				sP2 = *(gpsVert + (gpsEdge+iEdgeID2)->iPoint[1]);
				Extend_Points(&sP1, &sP2);
				if (asInter[3].fz > gsRS.fCeiling - 1)
					{ sP1.fz = gsRS.fCeiling;  sP2.fz = gsRS.fCeiling; }
				sPlane2 = Create_Edge_Plane(iEdgeID2, pCurrRay->sVirtTransLoc);
					/********/
				Break_Up_Ray_Edge(pCurrRay, asInter, sInterPlane, sPlane2, sP1, sP2, 1, asNewTri);
					/********/
				i=3;
			}
		}
	}
		/* set edge IDS in RAY_TRI_TYPE struct */
	for(i=0; i<3; i++) {		
		if (asNewTri[i].iActive == 1) {
			iYes2=1; iCount=i+1;
			if (asNewTri[i].iPlane==0)
				asNewTri[i].iEdgeID = iEdgeID;
			else
				asNewTri[i].iEdgeID = iEdgeID2;			
		}
	}
	if (iYes2 == 0 || iCount == 1)	return; /* degenerate case so do not split */

	for(i=0;i<4;i++)	aiStat[i]=0;

for(i=0; i<3; i++) {
	if (asNewTri[i].iActive == 1 && iLevel < 3) {
			/* check for ground reflection */
		iAll=0;
		if (asNewTri[i].iEdgeID != -2) {
			if (giCL_Terrain) { /* check to see if any points lie below the edge */
				sPlane = *(gpsTerrPlanes + pCurrRay->iTriID);
				iAll = 1;
				for(j=0; j<3; j++)
					if (Point_To_Plane_DistancePM(sPlane, asNewTri[i].asP[j]) < -0)	
						{ aiStat[i] = -2; }
					else	iAll = 0;
			}
			else {
				sPlane = gsGroundPlane;
				iAll = 1;
				for(j=0; j<3; j++) {
					if (asNewTri[i].asP[j].fz < -0)	
						{ aiStat[i] = -2; }
					else 	iAll = 0;
				}
			}
				/* check if all lie in the same triangle, if so, done */
			if (iAll && All_In_Triangle(asP, asNewTri[i].asP, *pCurrRay, sPlane)) {
				aiStat[i] = 0;
				for(j=0; j<4; j++)	asNewTri[i].asP[j] = asP[j];
				asNewTri[i].iEdgeID = -2;
			}
		}
/* printf(">>>> ground %i  %li %i\n", iAll, asNewTri[i].iEdgeID, aiStat[i]);*/
/*if (i==1) {Print_Point(asNewTri[i].asP[0]);
Print_Point(asNewTri[i].asP[1]);
Print_Point(asNewTri[i].asP[2]); }*
		/***************/
			/* check for a ceiling reflection */
		if (aiStat[i] == 0 && asNewTri[i].iEdgeID != -3)  {
			iAll = 1;
			for(j=0; j<3; j++)	
				if (asNewTri[i].asP[j].fz > gsRS.fCeiling)
					{ aiStat[i] = -3; }
				else	iAll = 0;
			if (iAll && All_In_Triangle(asP, asNewTri[i].asP, *pCurrRay, gsCeilingPlane)) {
				aiStat[i] = 0;
				for(j=0; j<4; j++)	asNewTri[i].asP[j] = asP[j];
				asNewTri[i].iEdgeID = -3;
			}
		}
/* printf(">>>> ceiling %i\n", iAll); */
		/******/
		if (aiStat[i] < 0) 
		{
	/* printf("second break %i %i %li\n", i, asNewTri[i].iPlane, asNewTri[i].iEdgeID);*/
			if (asNewTri[i].iPlane == 1)	sInterPlane = sPlane2;
				/******/
			Break_Up_Ray(pCurrRay, asNewTri[i].iEdgeID, sInterPlane, asNewTri[i].asP, aiStat, asNewTri+iCount, iLevel+1);
				/*****/
			iNewCount = iCount;
			for(k=0; k< MAX_NEW_TRI - ((iLevel+1) * 3); k++)
				if (asNewTri[k].iActive == 1)	iNewCount=k+1;
			if (iNewCount > iCount)	{ /* successfully split triangle */
				asNewTri[i].iActive = 0; /* ignore this triangle since it has been broken */
				iCount = iNewCount;
			}
		}
	}
} /* for i */
}

void Initialize_New_Rays(RAY_TRI_TYPE asNewTri[], short iNumTri, RAY_TYPE* pCurrRay, 
																	RAY_TYPE* psNewRays)
{
	short			i, j;
	PARA_LINE_TYPE	sLine;
	RAY_TYPE*		pNewRay = psNewRays;
	EDGE_TYPE*		psEdge = (EDGE_TYPE*) gpObst_Tri[1]; 
	
	for(i=0; i < iNumTri; i++) 
	{
		*pNewRay = *pCurrRay;
		pNewRay->sLKP = asNewTri[i].asP[3];
		pNewRay->iEdgeID = asNewTri[i].iEdgeID;
		for(j=0; j<3; j++) pNewRay->asTriP[j] = asNewTri[i].asP[j];
			/* if the new ray does not intersect the wall/ground/ceiling then compute new vectors */
		if (pNewRay->iEdgeID > 0 && (psEdge + pNewRay->iEdgeID)->iWallType == 0) 
		{
			pNewRay->iTriID = OtherTri(pCurrRay->iTriID, pNewRay->iEdgeID); 
			sLine = Line_Parametric_Equation3D(pNewRay->sVirtTransLoc, pNewRay->sLKP);
			pNewRay->sRayVect = sLine.sV;
			for(j=0; j<3; j++) {
				sLine = Line_Parametric_Equation3D(pNewRay->sVirtTransLoc, asNewTri[i].asP[j]);
				pNewRay->asTriVect[j] = sLine.sV;
			}
		}
		pNewRay++; 
	}
}	
	/*********************************************************************************
		Check_For_Split:
			The area of the end triangle of a ray-cone needs to be kept under a 
		  certain maximum size to maintain sufficient accuracy.  This procedure
		  checks for a size violation.
		  	asInter is the 4 intersection points the ray-cone makes with the plane
		of intersection. asInter[3] is the RayVect intersection point.
	**********************************************************************************/
int	Check_For_Split(POINT_TYPE asInter[], RAY_TYPE* pCurrRay)
{
	float			fArea;
	POINT_TYPE		asTriP[3];
	int				i, iSplit=0;

for(i=0; i<3; i++) {
	if (fabs(pCurrRay->asTriVect[i].i) < .05) 
		if (fabs(pCurrRay->asTriVect[0].i) > .1 || fabs(pCurrRay->asTriVect[1].i) > .1 || fabs(pCurrRay->asTriVect[2].i > .1))
			{ iSplit = 1; i = 3;}
	if (fabs(pCurrRay->asTriVect[i].j) < .05) 
		if (fabs(pCurrRay->asTriVect[0].j) > .1 || fabs(pCurrRay->asTriVect[1].j) > .1 || fabs(pCurrRay->asTriVect[2].j) > .1)
			{ iSplit = 1; i = 3;}
	if (fabs(pCurrRay->asTriVect[i].k) < .05) 
		if (fabs(pCurrRay->asTriVect[0].k) > .1 || fabs(pCurrRay->asTriVect[1].k) > .1 || fabs(pCurrRay->asTriVect[2].k > .1))
			{ iSplit = 1; i = 3;}
}
if (giCL_Split && !iSplit) {
	for(i=0; i<3; i++)	asTriP[i] = asInter[i];
		/***/
	fArea = Triangle_Area(asTriP);
		/* gfCL_Split is the square of the area of the parallelogram the triangle defines */
	if (fArea > gfCL_Split)	{

		if (giCL_Debug)	fprintf(gfp_Debug, "  >>Check_For_Split: Splitting Ray\n");
		iSplit = 1; 
	}			   
}
	return(iSplit);
}
	/*********************************************************************************
		Assign_Split_Vals:
	**********************************************************************************/
void Assign_Split_Vals(pCurrRay, pNewRay, sV0, sV1, sV2, sP0, sP1, sP2)
	RAY_TYPE*	pCurrRay, *pNewRay;
	VECTOR_TYPE	sV0, sV1, sV2;
	POINT_TYPE	sP0, sP1, sP2;
{
	PARA_LINE_TYPE	sL;

	*pNewRay = *pCurrRay;
		/* new vectors for new ray-cone */
	pNewRay->asTriVect[0] = sV0;
	pNewRay->asTriVect[1] = sV1; 
	pNewRay->asTriVect[2] = sV2; 
	pNewRay->asTriP[0] = sP0; pNewRay->asTriP[1] = sP1; pNewRay->asTriP[2] = sP2;
	pNewRay->sLKP = ComputeCentroid3D(pNewRay->asTriP); 
		/* calculate new RayVect for new ray-cone */
	sL = Line_Parametric_Equation3D(pNewRay->sVirtTransLoc, pNewRay->sLKP);
	pNewRay->sRayVect = sL.sV;
}  
	/*********************************************************************************
		Split_Ray:  This splits up the current ray-cone into 4 new ray-cones

			sInter: the previous intersection point computed in Walk_Until_Event (sLastInter)
			asTriP[]: the computed tetrahedron's triangular end face points
				used to calculate the midpoints for the split into 4 new triangles
	**********************************************************************************/
void Split_Ray(RAY_TYPE* pCurrRay, POINT_TYPE asTriP[], RAY_TYPE* psNewRays)
{
	POINT_TYPE		asM[3], sVirt = pCurrRay->sVirtTransLoc;
	PARA_LINE_TYPE	asL[3];
/* printf("splitting ray\n"); 
Print_Point(asTriP[0]);
Print_Point(asTriP[1]);
Print_Point(asTriP[2]); */
	asM[0] = MidPoint(asTriP[0], asTriP[1]);
	asM[1] = MidPoint(asTriP[1], asTriP[2]);
	asM[2] = MidPoint(asTriP[0], asTriP[2]);
	asL[0] = Line_Parametric_Equation3D(sVirt, asM[0]); 
	asL[1] = Line_Parametric_Equation3D(sVirt, asM[1]); 
	asL[2] = Line_Parametric_Equation3D(sVirt, asM[2]); 

		/***** Create the four new ray-cones *********/
			/* lower right triangle */
	Assign_Split_Vals(pCurrRay, (psNewRays+3), asL[2].sV, asL[1].sV, 
						pCurrRay->asTriVect[2], asM[2], asM[1], asTriP[2]);
			/* lower left triangle */
	Assign_Split_Vals(pCurrRay, (psNewRays+2), asL[0].sV, pCurrRay->asTriVect[1], 
						asL[1].sV, asM[0], asTriP[1], asM[1]);
			/* Upper triangle */
	Assign_Split_Vals(pCurrRay, (psNewRays+1), pCurrRay->asTriVect[0], asL[0].sV,
						asL[2].sV, asTriP[0], asM[0], asM[2]);
			/* middle triangle */
	Assign_Split_Vals(pCurrRay, psNewRays, asL[0].sV, asL[1].sV, asL[2].sV, 
						asM[0], asM[1], asM[2]);
}
	/*********************************************************************************
		0 - intersection with a wall
		1 - intersection with a vertex of the triangulation
		2 - intersection with the ground
		3 - intersection with the ceiling
		4 - Ray-cone's end face's area is greater than split threshold value
		5 - The ray-cone's power is below the beta threshold 
		6 - The ray-cone has propagated outside the bounding region
	**********************************************************************************/
short Event_Update(short aiEvents[], RAY_TYPE* pCurrRay, long iNextTri, long iEdgeID, 
					POINT_TYPE asInter[], long aiStatus[], RAY_TYPE* psNewRays)
{
	RAY_TRI_TYPE	asNewTri[MAX_NEW_TRI];
	RAY_TYPE*		pNewRay;
	short			i, iNumNew=0, iNumOld, iNumTri, iIndex;
	PLANE_TYPE		sInterPlane, sPlane;
	float			fDist, fNewRayPow;
	EDGE_TYPE*		psEdge = (EDGE_TYPE*) gpObst_Tri[1]; 

	if (iEdgeID > 0) 	sInterPlane = Create_Edge_Plane(iEdgeID, pCurrRay->sVirtTransLoc);
	else if (iEdgeID == -2)	
		if (giCL_Terrain)	sInterPlane = *(gpsTerrPlanes + pCurrRay->iTriID);
		else sInterPlane = gsGroundPlane;
	else	sInterPlane = gsCeilingPlane;
	
	if (giCL_Debug)	{ fprintf(gfp_Debug, "  EVENT_UPDATE:\n"); SaveDF(); }
			/******/
	fDist = Macro_Dist(asInter[3], pCurrRay->sVirtTransLoc);
		/* path_loss */
	fNewRayPow = Macro_Free_Space_Calc(pCurrRay->fAtten, fDist);

if (fNewRayPow > gsRS.fBeta)	
{	 		
	for(i=0; i<MAX_NEW_TRI; i++)	asNewTri[i].iActive = 0;
		/* must break up ray since ray-cone end intersects multiple faces of the prism */
	if (aiEvents[1] > 0) {
			/* two types of break ups: EdgePlane and ground/ceiling plane */
		Break_Up_Ray(pCurrRay, iEdgeID, sInterPlane, asInter, aiStatus, asNewTri, 0);
			/*****/
				/* 'condense' the results */
		iNumTri = 0;
		for(i=0; i<MAX_NEW_TRI; i++) 
			if (asNewTri[i].iActive == 1)	{ asNewTri[iNumTri] = asNewTri[i]; iNumTri++; }
/* Print_Ray_Tri(asNewTri, iNumTri); */
		/******/
		if (giCL_Debug) 	fprintf(gfp_Debug, "   Initializing newly broken rays %i \n", iNumNew); 
			/* initialize new rays */
		Initialize_New_Rays(asNewTri, iNumTri, pCurrRay, psNewRays);
			/********/
		iNumNew = iNumTri;
	}
	if (iNumNew == 0) { 	/* ray-cone was not broken */
		for(i=0; i<4; i++)	asNewTri[0].asP[i] = asInter[i];
		asNewTri[0].iActive=1;  asNewTri[0].iEdgeID = iEdgeID;
		iNumNew = 1;
		Initialize_New_Rays(asNewTri, 1, pCurrRay, psNewRays);
	}
/* printf("here12 NumNew %i   iNextTRi %li   iTriID %li   iEdgeID %li\n", iNumNew, iNextTri, pCurrRay->iTriID, pCurrRay->iEdgeID); */
		/* reflect and transmitt new ray-cones */
	for(i=0; i<iNumNew; i++)
	{
		pNewRay = (psNewRays + i); 
		iIndex = pNewRay->cNumEvent - '0';

		if (pNewRay->iEdgeID == -2) {	/* ground reflection */
/* printf("ground reflected %i  ", giCL_Terrain); */
			if (giCL_Terrain)
				sPlane = *(gpsTerrPlanes + pNewRay->iTriID);
			else	sPlane = gsGroundPlane;
				/* reflect and attenuate */
			Reflect_Ray(pNewRay, asNewTri[i].asP, sPlane);
			pNewRay->fAtten = Reflect_Attenuation(pNewRay->fAtten, 1, 0, Macro_DotProduct(pNewRay->sRayVect, sPlane.sN)); 
				/******/
					/* update ray's history */
			pNewRay->iEdgeID = From_Edge(*pNewRay, pNewRay->sLKP); 
			pNewRay->aiPath[iIndex] = pNewRay->iTriID;
				/* '0' - nothing '1' - wall reflection '2' - 'wall transmission' */
				/* '3' - ground reflection '4' - 'ceiling reflection' '5' - 'diffracted ray' */
			pNewRay->cPrevious = '3'; pNewRay->acEvent[iIndex] = '3'; pNewRay->cNumEvent++;
		}
		else if (pNewRay->iEdgeID == -3) {	/* ceiling reflection */
	/* printf("ceiling reflected  "); */
			if (giCL_InBuilding) {
				sPlane = gsCeilingPlane;
					/* reflect and attenuate */
				Reflect_Ray(pNewRay, asNewTri[i].asP, sPlane);
				pNewRay->fAtten = Reflect_Attenuation(pNewRay->fAtten, 2, 0, Macro_DotProduct(pNewRay->sRayVect, sPlane.sN)); 
					/*********/
				pNewRay->aiPath[iIndex] = -4;
				pNewRay->cPrevious = '4'; pNewRay->acEvent[iIndex] = '4'; pNewRay->cNumEvent++;
				pNewRay->iEdgeID = From_Edge(*pNewRay, pNewRay->sLKP); 
		/* printf("CEILING reflected\n"); cc=getchar(); */
			}
			else
				pNewRay->iTriID = -1;		/* ray-cone is leaving bounding region */
		}
		else if (pNewRay->iEdgeID < 0)
			{ printf("Error: iEdgeID in reflect < 0"); cc=getchar(); }
		else if ((psEdge + pNewRay->iEdgeID)->iWallType > 0) {	/* wall reflect */
 /*printf("reflected   "); cc=getchar();*/
			sPlane = Create_Edge_Plane(pNewRay->iEdgeID, pCurrRay->sVirtTransLoc);
				/* reflect and attenuate */
			Reflect_Ray(pNewRay, asNewTri[i].asP, sPlane);
			pNewRay->fAtten = Reflect_Attenuation(pNewRay->fAtten, (psEdge+pNewRay->iEdgeID)->iWallType, 1, Macro_DotProduct(pNewRay->sRayVect, sPlane.sN));
				/***** Update ray-cone's path history ****/
			pNewRay->aiPath[iIndex] = pNewRay->iEdgeID;
			pNewRay->cPrevious = '1'; pNewRay->acEvent[iIndex] = '1'; pNewRay->cNumEvent++;
		}
		else  { /* nothing: ray propagates onward */
/* printf("nothing %li %li  ", pNewRay->iEdgeID, pNewRay->iTriID); */
			pNewRay->cPrevious = '0';		
		}
	}	
	/* printf("\n"); */
} /* end break up and reflect/transmitt */
		/* check to see if any of the new ray-cones need to be split */

if (iNumNew > 0) {
	pNewRay = psNewRays; iNumOld = iNumNew;
	for(i=0; i<iNumOld; i++)
	{
		if (pNewRay->iTriID > 0 && Check_For_Split(asNewTri[i].asP, pNewRay)) {
			Split_Ray(pNewRay, asNewTri[i].asP, psNewRays+iNumNew);
			pNewRay->iTriID = -1; /* the split ray-cone can be discarded */
			iNumNew = iNumNew + 4;
		}
		pNewRay++;
	}
} /* end split */

	return(iNumNew);
}
	/*********************************************************************************
		Walk_Until_Event:
			First, check to see which triangle the ray would enter next.  Then
		check if any events would occur if the ray tried to enter the next
		triangle.  If any events would occur, stop the walk and return info.
		Otherwise, continue walking.
		0 - intersection with a wall
		1 - ray-cone contained in two faces of the prism
		2 - intersection with the ground
		3 - intersection with the ceiling
		4 - Ray-cone's end face's area is greater than split threshold value
		5 - The ray-cone's power is below the beta threshold 
		6 - The ray-cone has propagated outside the bounding region
			events 0-3 are not mutually exclusive with one another except for 2 and 3

		asInter[4]: stores the 4 ray-cone intersection points with the plane of incidence
			asInter[3] is sRayVects intersection point (centroid ray)
		aiStatus[4]: Event_Check returns status codes in this array
			-2 ground, -3 ceiling and >0 the triangle ID that the intersection point is in
	**********************************************************************************/
short Walk_Until_Event(RAY_TYPE* pCurrRay, long aiTriEncount[], RAY_TYPE* psNewRays)
{
	short		iNumTri;	/* tracks number of triangles encountered */
	short		iNumNew;	/* number of new ray-cones created - returned */
	short		iEvent, aiEvents[10], iRayDone=0, i;
	long		iNextTri=-2, iEdgeID, aiStatus[4];
	POINT_TYPE	asInter[4];
	
	iNumTri = 0; iNumNew=0; iEvent = 0;
	iEdgeID = pCurrRay->iEdgeID; 
	
	for(i=0; i<7; i++) 	aiEvents[i]=0;						 
	for(i=0; i<4; i++) 	aiStatus[i]=0; 

/*printf("walking ray %li\n", giNumRayStack); fflush(stdout);*/
/*if (giNumRayStack <= 451 )
	{ Print_Ray(*pCurrRay);   } */
	
	while (!iRayDone)
	{
			/* the triangle that is about to be 'walked' across */
		aiTriEncount[iNumTri] = pCurrRay->iTriID; iNumTri++;
			/* computes the next edge to walk over and the next triangle to walk into */
	iNextTri = Ray_TriEdge_Intersect(pCurrRay, &iEdgeID);
	/***/
		if (iNextTri == -1) /* ray has left the region defined by the scenario */
			{ aiEvents[6] = 1; iEvent=1; }
		else if (iNextTri == -4) /* Ray_TriEdge_intersection failed */
			{ giNumAbort++; aiEvents[6] = 1; iEvent=1; }
		else {
		Compute_Cone_End_Inter(pCurrRay, iEdgeID, asInter);
					/* check for the possible events */
		iEvent = Event_Check(aiEvents, pCurrRay, iNextTri, &iEdgeID, asInter, aiStatus);
		}
		/******/
			/* check if ray needs to be split as it walks - 
									only check if no other events have occured */
		if (iEvent == 0 && Check_For_Split(asInter, pCurrRay)) 
			{ iEvent = 1; aiEvents[4] = 1; }
		/****/
		if (giCL_Debug) { fprintf(gfp_Debug, "   %i aiEvent: %i %i %i %i %i %i %i\n", iEvent, aiEvents[0], aiEvents[1],aiEvents[2],aiEvents[3],aiEvents[4],aiEvents[5], aiEvents[6]); SaveDF(); }		
 /* printf("   %i aiEvent: %i %i %i %i %i %i %i\n", iEvent, aiEvents[0], aiEvents[1],aiEvents[2],aiEvents[3],aiEvents[4],aiEvents[5], aiEvents[6]); */
		/******/
		if (iEvent > 0 && aiEvents[6]==0) {
		iNumNew = Event_Update(aiEvents, pCurrRay, iNextTri, iEdgeID, asInter, aiStatus, psNewRays);
			iRayDone=1; 
		}
		else if (aiEvents[6] == 1)	/* ray leaving bounding region */
			iRayDone=1; 
		else {
			pCurrRay->iEdgeID = iEdgeID;
			pCurrRay->iTriID = iNextTri;
		}
	} /* while !RayDone */

	aiTriEncount[iNumTri] = -1;			/* -1 is the halting test */
		/* updated for trapezoid computation and RaySave.dat */
	for(i=0; i<3; i++)	pCurrRay->asTriP[i] = asInter[i];
	pCurrRay->sLKP = asInter[3];

	return(iNumNew);
}
	/*********************************************************************************
		Inside_Trap:
	  All of the trapezoids face normals point INTO the trapezoid.  Therefore
	Point_On_Plane should always be >=0 if the point lies to the 'inside' of the
	face.  If any of the tests are negative then the point lies outside of the trap.			
	**********************************************************************************/
short Inside_Trap(POINT_TYPE sLoc)
{
	short			iYes=1, i;

	for(i=0; i<5; i++) {
		if (Point_To_Plane_DistancePM(gsTrap.sP[i], sLoc) < 0)	
			{ iYes = 0; i=5; }
	}

	return(iYes);
}
	/*********************************************************************************
		QP_In_Triangles:
			loads the qpoints that are in the triangles that were encountered
		during the ray walk into an array, testing for containment in the tetra
		first.  It then creates a block of memory for the storage of these points 
		and returns a pointer to the block.
			sLKP - rays' last known position
	**********************************************************************************/
long* QP_In_Triangles(long aiTriEncount[])
{
	long	aiQP[1000], iTriID, iQP;	
	short	iCount=0, i, iQPcount=0;
	long*	pQP=NULL;

	while (aiTriEncount[iCount] != -1)
	{
		giNumWalked++;
		iTriID = aiTriEncount[iCount];	
		for(i=0; i<	(gpsTriQP+iTriID)->iNum; i++) {
			iQP = *((gpsTriQP+iTriID)->piTriQP + i);
			if (Inside_Trap((gpsQPoints+iQP)->sLoc)) {
				aiQP[iQPcount] = iQP;
				iQPcount++;
			}
		}
		iCount++;
	}
	if (iQPcount > 0) {
		pQP = (long*) calloc(iQPcount + 1, sizeof(long));
		for(i=0; i<iQPcount; i++)
			*(pQP+i) = aiQP[i];
		*(pQP+iQPcount) = -1;	/* marks the end of the list */
	}
	else
		pQP = NULL;

	return(pQP);
}
	/*********************************************************************************
		Saves the path for each captured query point in RayPaths.dat
	**********************************************************************************/
void Save_Query_Points_Path(long aiUpdateQP[], RAY_TYPE* pCurrRay, short aiQP_Hit[])
{
	short		iCount=0, j;
	long		iQP_ID, i;
	float		fDist, fDistR;
	
	while (aiUpdateQP[iCount] != -1)		iCount++;

	for(i=0; i<iCount; i++) {
		giNumSavedPaths++;
		iQP_ID = aiUpdateQP[i];
		aiQP_Hit[iQP_ID]++;
		fDist = Macro_Dist((gpsQPoints + iQP_ID)->sLoc, pCurrRay->sVirtTransLoc);
		fDistR = (float) sqrt(fDist);

		fprintf(gfpSave_Paths, "%li %c\n", iQP_ID, pCurrRay->cNumEvent);
		for(j=0; j<MAX_EVENTS; j++)
			fprintf(gfpSave_Paths, "%c ", pCurrRay->acEvent[j]);
		fprintf(gfpSave_Paths, "\n");
		for(j=0; j<MAX_EVENTS; j++)
			fprintf(gfpSave_Paths, "%li ", pCurrRay->aiPath[j]);
		fprintf(gfpSave_Paths, "\n");
		fprintf(gfpSave_Paths, "%f %f %f\n", pCurrRay->sLKP.fx, pCurrRay->sLKP.fy, pCurrRay->sLKP.fz);
		fprintf(gfpSave_Paths, "%f %f\n", fDistR, pCurrRay->fAtten);
	} 
}

	/*********************************************************************************
		Update_RayStack: 
			Adds the new rays onto the ray stack and removes the top
	ray from the stack (it is done).  As a ray dies it is saved to the global file
	'RaySave.dat' if the giCL_RaySave variable is != 0.
		- Returns a pointer to the top of the stack
		- updates the gloabl ray stack variable, giNumRayStack
	**********************************************************************************/
RAY_TYPE* Update_RayStack(RAY_TYPE asNewRays[], short iNumRays, RAY_TYPE sCurrRay)
{
	short		i;
	RAY_TYPE*	pRays=gpsRayStack;

	giNumRayStack--; pRays--;

		/* add new rays, if any */
	for(i=0; i<iNumRays; i++) {
		if (asNewRays[i].iTriID >= 0 && asNewRays[i].cNumEvent <= gsRS.cMaxEvents 
			&& asNewRays[i].sLKP.fx > -.01 && asNewRays[i].sLKP.fy > -.01 
			&& Triangle_Area(asNewRays[i].asTriP) > .0025) {
		Check_Ray_Cone(asNewRays[i], (asNewRays + i));
			if (asNewRays[i].iTriID >= 0) {
				pRays++; giNumRayStack++;
				*pRays = asNewRays[i];
			}
		}
/*		else if (asNewRays[i].sLKP.fx < -.01 && asNewRays[i].sLKP.fy < -.01)
		{
			Print_Ray(asNewRays[i]);
			cc=getchar();
		}*/
	}
		
	if (giCL_Debug)	{
		fprintf(gfp_Debug, "   SIZE of RAYSTACK: %li\n", giNumRayStack);
		SaveDF();
	}

	return(pRays);
}
void Save_Ray(RAY_TYPE sO, RAY_TYPE sF)
{
	int		i;

	if (giNumRayStack == giInitRayNum) {
		if (gfpRay == NULL) {
			gfpRay = fopen("RaySave.dat", "a+");
			fprintf(gfpRay, "BEGIN %li\n", giNumRayStack);
		}
		else {
			gfpRay = fopen("RaySave.dat", "a+");
			fprintf(gfpRay, "END %li\n", giNumRayStack + 1);
			fprintf(gfpRay, "BEGIN %li\n", giNumRayStack);
		}
		giInitRayNum--;
	}
	else
		gfpRay = fopen("RaySave.dat", "a+");

	fprintf(gfpRay, "%f %f %f %f %f %f\n", sO.sLKP.fx, sO.sLKP.fy, sO.sLKP.fz,
												sF.sLKP.fx, sF.sLKP.fy, sF.sLKP.fz);
	for(i=0; i<3; i++)
		fprintf(gfpRay, "%f %f %f ", sO.asTriP[i].fx, sO.asTriP[i].fy, sO.asTriP[i].fz);
	fprintf(gfpRay, "\n");
	for(i=0; i<3; i++)
		fprintf(gfpRay, "%f %f %f ", sF.asTriP[i].fx, sF.asTriP[i].fy, sF.asTriP[i].fz);
	fprintf(gfpRay, "\n");
}
void SaveQP_To_RayFile(long aiUpdateQP[])
{
	short		i, iCount=0;
	QP_TYPE*	pQP;
	
	if (aiUpdateQP != NULL) {
		while (aiUpdateQP[iCount] != -1)		iCount++;

		fprintf(gfpRay, "%i ", iCount);
		for(i=0; i<iCount; i++) {
			pQP = (gpsQPoints + aiUpdateQP[i]);
			fprintf(gfpRay, "%f %f %f ", pQP->sLoc.fx, pQP->sLoc.fy, pQP->sLoc.fz); 
		}
		fprintf(gfpRay, "\n");
		fclose(gfpRay);
	}
	else 
		{ fprintf(gfpRay, "0 \n"); fclose(gfpRay); }
	gfpRay = NULL;
}
	/*********************************************************************************
		Walk_Ray:
			aiTriEncount - will be a pointer to a list of the triangles encountered by
					the ray as it walked to an event.  I say encounter because
					triangles not on the walk may need to be checked as well because
					it is a cone not a ray.  The cone may fall outside of the walked
					triangle.
			piUpdateQP - will be a pointer to the list of query point ID #s that
				need to be checked for containment within the ray-cone.
	**********************************************************************************/
RAY_TYPE* Walk_Ray(RAY_TYPE* pCurrRay, short aiQP_Hit[])
{
	long		aiTriEncount[500];	/* very doubtful if the ray would encounter more 
										than 100 triangles before an event */
	long*		piUpdateQP;		
	RAY_TYPE	*pStackTop, asNewRays[MAX_NEW_TRI];	
	short		iNumRays;
	RAY_TYPE	sOrigRay = *pCurrRay;

	
		/* returns a list of new rays created during an event (NULL if ray is finished) */
		/* also returns the list of triangles that have been encountered by the ray */
	iNumRays = Walk_Until_Event(pCurrRay, aiTriEncount, asNewRays);
		/* returns a list of query points that have to be updated */
	if (aiTriEncount[0] != -1) {
				/* compute the capture cone */
		gsTrap = Compute_Trap(pCurrRay->asTriP, sOrigRay.asTriP, (short) (pCurrRay->cNumEvent - '0'));	
		piUpdateQP = QP_In_Triangles(aiTriEncount);
			/* update the power at a query point */
		if (piUpdateQP != NULL) 	Save_Query_Points_Path(piUpdateQP, &sOrigRay, aiQP_Hit); 
	}
	pStackTop = Update_RayStack(asNewRays, iNumRays, *pCurrRay);
	
	if (giCL_RaySave) 	{ Save_Ray(sOrigRay, *pCurrRay); SaveQP_To_RayFile(piUpdateQP); }

	if (piUpdateQP != NULL) free(piUpdateQP);

	return(pStackTop);	/* add the new rays and return pointer to top */
}

	/*********************************************************************************
		Compute_PathLosses:
	**********************************************************************************/
short Compute_Paths(RAY_TYPE asRayStack[], short aiQP_Hit[])
{
		/* this initially loads the stack with the rays 
			shot from the transmitter at the specified angles */
	giNumRayStack = Initialize_Ray_Stack(asRayStack); 

		/* Jump to a specific ray top be traced - for debugging purposes */
 	/* giNumRayStack = giNumRayStack - 450;
	gpsRayStack = gpsRayStack - 450; */

	while (giNumRayStack > 0)
	{
		giNumRaysTraced++;		/* tracks the number of rays 'walked' */
		if (giCL_Debug) { fprintf(gfp_Debug, "\nWORKING ON RAY: %li\n", giNumRayStack); SaveDF(); }
			/* a ray is walked until it dies or an event occurs (ex: hits a wall) */
			/* Qp updating occurs during the walk */
		gpsRayStack = Walk_Ray(gpsRayStack, aiQP_Hit);	
	}
	printf("   Number of Rays Traced: %i\n", giNumRaysTraced);
	if (giCL_RaySave) {
		gfpRay = fopen("RaySave.dat", "a+");
		fprintf(gfpRay, "END 1\n");
		fclose(gfpRay);
	}

	return(1);
}

/***************************************************************************************
								STORE RESULTS
***************************************************************************************/
int Duplicate_Path(PATH_INF_TYPE asPaths[], short iNP)
{
	short 	i, k, iRet=0;
	
	for(i=0; i<iNP; i++)
	{
		iRet = 1;
		for(k=0; k<MAX_EVENTS; k++) 
			if (asPaths[iNP].acEvent[k] != asPaths[i].acEvent[k])	{ iRet=0; k=MAX_EVENTS; }
		if (iRet == 1)
			for(k=0; k<MAX_EVENTS; k++) 
				if (asPaths[iNP].aiPath[k] != asPaths[i].aiPath[k])	{ iRet=0; k=MAX_EVENTS; }
	}

	return(iRet);
}
QP_RAY_STORE_TYPE Compute_MultiPath_Signal_Strength(PATH_INF_TYPE asPaths[], 
															short iNP, float fTransPower)
{
	QP_RAY_STORE_TYPE	sQP_Store;
	int					i;
	float				fPow, fShort=(float) 10000, fD2;

	sQP_Store.fSignalPow = (float) 0; sQP_Store.fLongDist = (float) 0; 
	sQP_Store.fShortDist = (float) 1000; sQP_Store.fMaxPow = (float) 0;

	for(i=0; i<iNP; i++)
		if (fShort > asPaths[i].fDist)		fShort = asPaths[i].fDist;

	for(i=0; i<iNP; i++) {
		if (asPaths[i].fDist < fShort + 240) {
			fD2 = asPaths[i].fDist * asPaths[i].fDist;
			fPow = Macro_Free_Space_Calc(fTransPower * asPaths[i].fAtten, fD2);
			sQP_Store.fSignalPow = sQP_Store.fSignalPow + fPow;
			if (fPow > sQP_Store.fMaxPow)
				{ sQP_Store.fMaxPow = fPow; sQP_Store.fMaxDist = asPaths[i].fDist; }
			if (asPaths[i].fDist < sQP_Store.fShortDist)
				{ sQP_Store.fShortDist = asPaths[i].fDist; sQP_Store.fShortPow = fPow; }
			if (asPaths[i].fDist > sQP_Store.fLongDist)
				{ sQP_Store.fLongDist = asPaths[i].fDist; sQP_Store.fLongPow = fPow; }
		}
	}
	return(sQP_Store);
}
short Store_Received_Signal_Strengths(short aiQP_Hit[])
{
	QP_RAY_STORE_TYPE 	sQP_Store;
	long				i, j, iQP_ID, iC=0, iC2=0;
	PATH_INF_TYPE		asPaths[50];
	FILE*				fp;
	short				iNP, k;

	fclose(gfpSave_Paths);
	fp = fopen("RayPaths.dat", "r");
	fprintf(gfp_Paths, "BEGIN TRANSMITTER %i\n", gsRS.iTransID);
for(i=0; i<giNumQP; i++) {
	if (aiQP_Hit[i] > 0) 
	{
		iNP = 0;
		for(j=0; j<giNumSavedPaths; j++) {

			fscanf(fp, "%li %c\n", &iQP_ID, &asPaths[iNP].cNumEvent);
			for(k=0; k<MAX_EVENTS; k++) 
				fscanf(fp, "%c ", &asPaths[iNP].acEvent[k]); 
			for(k=0; k<MAX_EVENTS; k++) 
				fscanf(fp, "%li ", &asPaths[iNP].aiPath[k]);
			fscanf(fp, "%f %f %f", &asPaths[iNP].sLKP.fx, &asPaths[iNP].sLKP.fy, &asPaths[iNP].sLKP.fz);
			fscanf(fp, "%f %f", &asPaths[iNP].fDist, &asPaths[iNP].fAtten);

			if (iQP_ID == i) 
				if (!Duplicate_Path(asPaths, iNP))	iNP++;
			if (iNP == aiQP_Hit[i] || iNP==50)		j = giNumSavedPaths;
		}
		iC2 = iC2 + iNP;
		rewind(fp);
		if (iNP == 0)	{ printf("No paths found for Query Point %i in Store_Received_Signal_Strength\n", i); cc=getchar(); }

		fprintf(gfp_Paths, "BEGIN QUERY POINT %li %i   %f %f %f\n", i, iNP, (gpsQPoints + i)->sLoc.fx, (gpsQPoints + i)->sLoc.fy, (gpsQPoints + i)->sLoc.fz);
		for(j=0; j<iNP; j++) {
			fprintf(gfp_Paths, "%c\n", asPaths[j].cNumEvent);
			for(k=0; k<MAX_EVENTS; k++) 
				fprintf(gfp_Paths, "%c ", asPaths[j].acEvent[k]);
			fprintf(gfp_Paths, "\n");
			for(k=0; k<MAX_EVENTS; k++) 
				fprintf(gfp_Paths, "%li ", asPaths[j].aiPath[k]);
			fprintf(gfp_Paths, "\n");
			fprintf(gfp_Paths, "%f %f %f\n", asPaths[j].sLKP.fx, asPaths[j].sLKP.fy, asPaths[j].sLKP.fz);
			fprintf(gfp_Paths, "%f %f\n", asPaths[j].fDist, asPaths[j].fAtten);
		}
		fprintf(gfp_Paths, "END QUERY POINT: %li\n", i);

		sQP_Store = Compute_MultiPath_Signal_Strength(asPaths, iNP, (gpsTrans_Data+gsRS.iTransID)->fMax_Power);
  		sQP_Store.iQP_ID = i;

		if (sQP_Store.fSignalPow > 0) 
		{ 
			iC++;
			Add_Signal_Data((gpsQPoints+sQP_Store.iQP_ID), *(gpsTrans_Data+gsRS.iTransID), 
				sQP_Store.fSignalPow, sQP_Store.fShortDist, sQP_Store.fShortPow,
				sQP_Store.fLongDist, sQP_Store.fLongPow);	 
		} 
	}
} /* for i */
	fprintf(gfp_Paths, "END TRANSMITTER %i\n", gsRS.iTransID);
	fclose(fp);
	if (giCL_Stats) {
		printf("   Number of Query Points with a signal > Beta-Threshold: %i\n", iC); 
		if (iC > 0) printf("        Average number of paths per captured query point: %f\n", (float) iC2 / iC); 
	}
	return(1);
}

void Initialize_Tracking_Stuff(short iTransID) 
{
		/* track number of rays that were aborted due to problems */
	giNumAbort = 0;
		/* track number of time a degenerate intersection condition occurs */
	giNumInter = 0;
		/* track number of triangles walked */
	giNumWalked = 0;
		/* maximum allowable size of the ray stack */
	giRayMax = (int) (20 * pow(4, giCL_recursion_level) * 2 + 100);
		/* tracks the number of paths saved into RayPaths.dat */
	giNumSavedPaths = 0;
		/* open the file to save the ray paths in */
	if (giCL_RaySave) {
		gfpRay = NULL;
		if (iTransID == 0) {
			gfpRay = fopen("RaySave.dat", "w");
			if (gfpRay == NULL)	ERROR_Message("SAVE_RAY", "could not open RaySave.dat");
			fprintf(gfpRay, "%f %f %f %f %f %f\n", gasBR[0].fx, gasBR[0].fy, gasBR[0].fz, gasBR[1].fx, gasBR[1].fy, gasBR[1].fz);
		}
	}
}
/***************************************************************************************
				THE MAIN FUNCTION: external function that is called
***************************************************************************************/

/***************************************************************************************
	Ray_Method:  called by the outside system.  Sets up and then runs the ray-launching
			modeling method.
	 Input:
		aiTransID[] - an array of ID #s for each transmitter predictions are to be 
			performed upon
		iMaxReflect - maximum number of reflections a ray is allowed to make
		fBeta - the signal power cutoff threshold.  A ray is finished being traced
			either when it has reached the max # of reflections or when its power
			is below fBeta.
		iPropType - for selecting propagation formula to be used
		iAttenType - for selecting attenuation formula to be used
		iReflectType - 0 - angle of incidence = angle of reflection
					   1 - table lookup
					   2 - based upon electromagnetic wave properties of the material
	  Creates:
	  	asQP_Store[# Query Points] - used to store the rays that intersect the query 
	  					points
		gpsTriQP[# of Triangles] - used to store which query points a triangle contains
			if (gpsTriQP+x)->iNum == 0 then no query points are in the triangle.
			Otherwise, it is a pointer to the list of Query Point ID #s where the
			first entry is the number contained in the list.
***************************************************************************************/
short Ray_Method(short iTransID, short iMaxEvents, double fBeta, short iPropForm, 
																short iReflTransForm)
{
  	short*				piQP_Hit = (short*) calloc(giNumQP, sizeof(short));
	RAY_TYPE*			asRayStack = calloc((unsigned int) (20 * pow(4, giCL_recursion_level) * 2 + 100), sizeof(RAY_TYPE));
	int					iRet_Code; 

	if (asRayStack == NULL) {
		printf("Not enough memory to create ray stack.  Use the n# command\n   line option to set a lower number of initial rays to be launched.\n");
		cc=getchar();
		return(0);
	}
	printf("TRANSMITTER ID: %i\n", iTransID); fflush(stdout);
	gsRS.iTransID = iTransID;  
	gsRS.iNumQP = (gpsTrans_Data+iTransID)->iNum_Points;
	gsRS.piQP = (gpsTrans_Data+iTransID)->piQPoints;
	
	gsRS.cMaxEvents = '0' + iMaxEvents; gsRS.fBeta = fBeta;
	gsRS.iPropForm = iPropForm;
	gsRS.iReflTransForm = iReflTransForm; gsRS.iNumRecurse = giCL_recursion_level;
			/* Polarization: 0 - vertical  1 - horizontal */
	gsRS.iPolarization = (gpsTrans_Data+iTransID)->iPolarization;	

	gpsRayStack = NULL;		/* initialize global ray stack variables */
	giNumRayStack = 0; 
	giNumRaysTraced = 0;	/* tracks number of rays walked */

		/* it is assumed here that the height of the transmitter is taller than
			all receiver heights.  This can be relaxed with some extra code but
			is a reasonable assumption.  Thus, the transmitter height is the
			ceiling height for rays that 'escape' above the ceiling */
	if (giCL_InBuilding)
		gsRS.fCeiling = gfFloorHeight;
	else
		gsRS.fCeiling = (gpsTrans_Data+iTransID)->sLoc.fz + 5;
		/**************************************/
	gfpSave_Paths = fopen("RayPaths.dat", "w");
		/**************************************/
	Initialize_Ray_Data_Structures(piQP_Hit);
		/* initializes files for tracking errors, rays and other data */
	Initialize_Tracking_Stuff(iTransID);

	iRet_Code = Compute_Paths(asRayStack, piQP_Hit);
	iRet_Code = Store_Received_Signal_Strengths(piQP_Hit);

	/*****************/
	if (giCL_Stats) printf("   Number of Rays Aborted: %i     #Degenerate: %i\n", giNumAbort, giNumInter);
	if (giCL_Stats) printf("   Number of triangles walked: %li\n", giNumWalked);

	printf("\nEND TRANSMITTER: %i\n\n", iTransID);

	return(iRet_Code);
}
